//+------------------------------------------------------------------+
//|                                                   SignalCMLP.mqh |
//|                   Copyright 2009-2017, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#include <Expert\ExpertSignal.mqh>
#include <My\Cmlp.mqh>
//+------------------------------------------------------------------+
// wizard description start
//+------------------------------------------------------------------+
//| Description of the class                                         |
//| Title=Signals based on Normalization for Feature Scaling.        |
//| Type=SignalAdvanced                                              |
//| Name=CMLP                                                        |
//| ShortName=CMLP                                                   |
//| Class=CSignalCMLP                                                |
//| Page=signal_cgan                                                 |
//| Parameter=TrainSet,int,260,Training Set Size                     |
//| Parameter=Epochs,int,65,Epochs                                   |
//| Parameter=LearningRate,double,0.0005,Learning Rate               |
//| Parameter=DecayEpochSteps,int,5,Decay Epoch Steps                |
//| Parameter=DecayRateA,double,0.5,Decay Rate A                     |
//| Parameter=DecayRateB,double,0.5,Decay Rate B                     |
//| Parameter=PolynomialPower,int,2,Polynomial Power                 |
//| Parameter=MinLearningRate,double,0.0001,Minimum Learning Rate    |
//| Parameter=DefaultWeight,double,0.005,Default Weight              |
//| Parameter=DefaultBias,double,0.0,Default Bias                    |
//| Parameter=NormBeta_H_1,double,0.0,Normalization Beta Hidden 1    |
//| Parameter=NormBeta_H_2,double,0.0,Normalization Beta Hidden 2    |
//| Parameter=NormBeta_H_3,double,0.0,Normalization Beta Hidden 3    |
//| Parameter=NormBeta_H_4,double,0.0,Normalization Beta Hidden 4    |
//| Parameter=NormBeta_H_5,double,0.0,Normalization Beta Hidden 5    |
//| Parameter=NormBeta_H_6,double,0.0,Normalization Beta Hidden 6    |
//| Parameter=NormBeta_H_7,double,0.0,Normalization Beta Hidden 7    |
//| Parameter=NormBeta_H_8,double,0.0,Normalization Beta Hidden 8    |
//| Parameter=NormGamma_H_1,double,0.0,Normalization Gamma Hidden 1  |
//| Parameter=NormGamma_H_2,double,0.0,Normalization Gamma Hidden 2  |
//| Parameter=NormGamma_H_3,double,0.0,Normalization Gamma Hidden 3  |
//| Parameter=NormGamma_H_4,double,0.0,Normalization Gamma Hidden 4  |
//| Parameter=NormGamma_H_5,double,0.0,Normalization Gamma Hidden 5  |
//| Parameter=NormGamma_H_6,double,0.0,Normalization Gamma Hidden 6  |
//| Parameter=NormGamma_H_7,double,0.0,Normalization Gamma Hidden 7  |
//| Parameter=NormGamma_H_8,double,0.0,Normalization Gamma Hidden 8  |
//| Parameter=NormBeta_O_1,double,0.0,Normalization Beta Output 1    |
//| Parameter=NormGamma_O_1,double,0.0,Normalization Gamma Output 1  |
//+------------------------------------------------------------------+
// wizard description end
//+------------------------------------------------------------------+
//| Class CSignalCMLP.                                               |
//| Purpose: Class of MLP with Normalization for Feature Scaling.    |
//|            Derives from class CExpertSignal.                     |
//+------------------------------------------------------------------+
enum Elearning
{  LEARNING_FIXED = 0,
   LEARNING_STEP_DECAY = 1,
   LEARNING_EXPONENTIAL_DECAY = 2,
   LEARNING_POLYNOMIAL_DECAY = 3,
   LEARNING_INVERSE_TIME_DECAY = 4,
   LEARNING_COSINE_ANNEALING = 5,
   LEARNING_CYCLICAL = 6,
   LEARNING_ONE_CYCLE = 7,
   LEARNING_ADAPTIVE = 8
};
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
#define  __MLP_INPUTS 5
#define  __LAYERS 1
#define  __SIZES 8
#define  __MLP_OUTPUTS 1
int      __MLP_SETTINGS[2 + __LAYERS] // network settings

=
{  __MLP_INPUTS,  __SIZES,    __MLP_OUTPUTS
};
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CSignalCMLP   : public CExpertSignal
{
protected:

   int                           m_train_set;                       //
   int                           m_epochs;                              // Epochs
   double                        m_learning_rate;                       // Epsilon

   Elearning                     m_learning_type;

   double                        m_train_deposit;
   double                        m_train_criteria;

   int                           m_decay_epoch_steps;
   double                        m_decay_rate_a;
   double                        m_decay_rate_b;
   int                           m_polynomial_power;
   double                        m_min_learning_rate;                       // Epsilon
   Eadaptive                     m_adaptive_type;
                          
   Enormalize                    m_normalize_type;
   
   double                        m_norm_beta_h_1;
   double                        m_norm_beta_h_2;
   double                        m_norm_beta_h_3;
   double                        m_norm_beta_h_4;
   double                        m_norm_beta_h_5;
   double                        m_norm_beta_h_6;
   double                        m_norm_beta_h_7;
   double                        m_norm_beta_h_8;
   double                        m_norm_beta_o_1;
   
   double                        m_norm_gamma_h_1;
   double                        m_norm_gamma_h_2;
   double                        m_norm_gamma_h_3;
   double                        m_norm_gamma_h_4;
   double                        m_norm_gamma_h_5;
   double                        m_norm_gamma_h_6;
   double                        m_norm_gamma_h_7;
   double                        m_norm_gamma_h_8;
   double                        m_norm_gamma_o_1;

   ENUM_ACTIVATION_FUNCTION      m_activation;
   ENUM_LOSS_FUNCTION            m_loss;

   double                        m_prior_learning_rate;

   double                        m_default_weight;
   double                        m_default_bias;
   
public:
                     CSignalCMLP(void);
                    ~CSignalCMLP(void);

   //--- methods of setting adjustable parameters
   void                          TrainSet(int value)
   {  m_train_set = value;
   }
   void                          Epochs(int value)
   {  m_epochs = value;
   }
   void                          LearningRate(double value)
   {  m_learning_rate = value;
   }
   void                          LearningType(Elearning value)
   {  m_learning_type = value;
   }
   void                          DecayEpochSteps(int value)
   {  m_decay_epoch_steps = value;
   }
   void                          DecayRateA(double value)
   {  m_decay_rate_a = value;
   }
   void                          DecayRateB(double value)
   {  m_decay_rate_b = value;
   }
   void                          PolynomialPower(int value)
   {  m_polynomial_power = value;
   }
   void                          MinLearningRate(double value)
   {  m_min_learning_rate = value;
   }
   void                          AdaptiveType(Eadaptive value)
   {  m_adaptive_type = value;
   }
   void                          NormalizeType(Enormalize value)
   {  m_normalize_type = value;
   }
   void                          NormBeta_H_1(double value)
   {  m_norm_beta_h_1 = value;
   }
   void                          NormBeta_H_2(double value)
   {  m_norm_beta_h_2 = value;
   }
   void                          NormBeta_H_3(double value)
   {  m_norm_beta_h_3 = value;
   }
   void                          NormBeta_H_4(double value)
   {  m_norm_beta_h_4 = value;
   }
   void                          NormBeta_H_5(double value)
   {  m_norm_beta_h_5 = value;
   }
   void                          NormBeta_H_6(double value)
   {  m_norm_beta_h_6 = value;
   }
   void                          NormBeta_H_7(double value)
   {  m_norm_beta_h_7 = value;
   }
   void                          NormBeta_H_8(double value)
   {  m_norm_beta_h_8 = value;
   }
   void                          NormGamma_H_1(double value)
   {  m_norm_gamma_h_1 = value;
   }
   void                          NormGamma_H_2(double value)
   {  m_norm_gamma_h_2 = value;
   }
   void                          NormGamma_H_3(double value)
   {  m_norm_gamma_h_3 = value;
   }
   void                          NormGamma_H_4(double value)
   {  m_norm_gamma_h_4 = value;
   }
   void                          NormGamma_H_5(double value)
   {  m_norm_gamma_h_5 = value;
   }
   void                          NormGamma_H_6(double value)
   {  m_norm_gamma_h_6 = value;
   }
   void                          NormGamma_H_7(double value)
   {  m_norm_gamma_h_7 = value;
   }
   void                          NormGamma_H_8(double value)
   {  m_norm_gamma_h_8 = value;
   }
   void                          NormBeta_O_1(double value)
   {  m_norm_beta_o_1 = value;
   }
   void                          NormGamma_O_1(double value)
   {  m_norm_gamma_o_1 = value;
   }
   void                          ActivationType(ENUM_ACTIVATION_FUNCTION value)
   {  m_activation = value;
   }
   void                          LossType(ENUM_LOSS_FUNCTION value)
   {  m_loss = value;
   }
   void                          DefaultWeight(double value)
   {  m_default_weight = value;
   }
   void                          DefaultBias(double value)
   {  m_default_bias = value;
   }

   //--- method of verification of settings
   virtual bool      ValidationSettings(void);
   //--- method of creating the indicator and timeseries
   virtual bool      InitIndicators(CIndicators *indicators);
   //--- methods of checking if the market models are formed
   virtual int       LongCondition(void);
   virtual int       ShortCondition(void);

protected:
   void                          GetOutput(double &Output);
   void                          SetOutput();

   Cmlp                          *MLP;
   
   void              Normalize(vector &Data);
};
//+------------------------------------------------------------------+
//| Constructor                                                      |
//+------------------------------------------------------------------+
void CSignalCMLP::CSignalCMLP(void) :    m_train_set(260),
   m_epochs(30),
   m_learning_rate(0.0005),
   m_learning_type(LEARNING_ADAPTIVE),
   m_decay_epoch_steps(5),
   m_decay_rate_a(0.5),
   m_decay_rate_b(0.5),
   m_polynomial_power(2),
   m_min_learning_rate(0.0001),
   m_adaptive_type(ADAPTIVE_GRADIENT),
   m_normalize_type(NORMALIZE_FEATURE),
   m_activation(AF_SIGMOID)

{
//--- initialization of protected data
   m_used_series = USE_SERIES_OPEN + USE_SERIES_HIGH + USE_SERIES_LOW + USE_SERIES_CLOSE + USE_SERIES_SPREAD + USE_SERIES_TIME;
   m_prior_learning_rate = m_learning_rate;
   vector _beta[],_gamma[];
   ArrayResize(_beta, __LAYERS + 1);
   ArrayResize(_gamma, __LAYERS + 1);
   _beta[0].Init(__MLP_SETTINGS[1]);
   _beta[1].Init(__MLP_SETTINGS[2]);
   _gamma[0].Init(__MLP_SETTINGS[1]);
   _gamma[1].Init(__MLP_SETTINGS[2]);
   _beta[0][0] = m_norm_beta_h_1;
   _beta[0][1] = m_norm_beta_h_2;
   _beta[0][2] = m_norm_beta_h_3;
   _beta[0][3] = m_norm_beta_h_4;
   _beta[0][4] = m_norm_beta_h_5;
   _beta[0][5] = m_norm_beta_h_6;
   _beta[0][6] = m_norm_beta_h_7;
   _beta[0][7] = m_norm_beta_h_8;
   _beta[1][0] = m_norm_beta_o_1;
   _gamma[0][0] = m_norm_gamma_h_1;
   _gamma[0][1] = m_norm_gamma_h_2;
   _gamma[0][2] = m_norm_gamma_h_3;
   _gamma[0][3] = m_norm_gamma_h_4;
   _gamma[0][4] = m_norm_gamma_h_5;
   _gamma[0][5] = m_norm_gamma_h_6;
   _gamma[0][6] = m_norm_gamma_h_7;
   _gamma[0][7] = m_norm_gamma_h_8;
   _gamma[1][0] = m_norm_gamma_o_1;
   MLP = new Cmlp(__MLP_SETTINGS, m_activation, m_loss, _beta, _gamma, m_normalize_type, m_default_weight, m_default_bias, m_min_learning_rate);
   m_train_deposit = AccountInfoDouble(ACCOUNT_BALANCE);
   m_train_criteria = 0.0;
   m_prior_learning_rate = m_learning_rate;
}
//+------------------------------------------------------------------+
//| Destructor                                                       |
//+------------------------------------------------------------------+
void CSignalCMLP::~CSignalCMLP(void)
{  delete MLP;
}
//+------------------------------------------------------------------+
//| Validation settings protected data.                              |
//+------------------------------------------------------------------+
bool CSignalCMLP::ValidationSettings(void)
{  if(!CExpertSignal::ValidationSettings())
      return(false);
//--- initial data checks
   if(m_epochs < 1)
   {  printf(__FUNCSIG__ + " epochs size should at least be 1! ");
      return(false);
   }
   if(m_learning_rate < 0.0)
   {  printf(__FUNCSIG__ + " learning rate should be greater than 0.0! ");
      return(false);
   }
   if(!MLP.validated)
   {  printf(__FUNCSIG__ + " invalid network settings. ");
      return(false);
   }
//read best weights
//datetime _version;
//N.Get(m_symbol.Name()+"_"+EnumToString(m_period), m_train_criteria, _version);
//--- ok
   return(true);
}
//+------------------------------------------------------------------+
//| Create indicators.                                               |
//+------------------------------------------------------------------+
bool CSignalCMLP::InitIndicators(CIndicators *indicators)
{
//--- check pointer
   if(indicators == NULL)
      return(false);
//--- initialization of indicators and timeseries of additional filters
   if(!CExpertSignal::InitIndicators(indicators))
      return(false);
//--- ok
   return(true);
}
//+------------------------------------------------------------------+
//| "Voting" that price will grow.                                   |
//+------------------------------------------------------------------+
int CSignalCMLP::LongCondition(void)
{  int result = 0;
   double _mlp_output = 0.0;
   GetOutput(_mlp_output);
   if(_mlp_output > 0.5)
   {  vector _new, _old;
      _new.CopyRates(m_symbol.Name(), m_period, 8, 0, __MLP_INPUTS);
      _old.CopyRates(m_symbol.Name(), m_period, 8, 1, __MLP_INPUTS);
      result = int(round(100.0 * (fabs(_mlp_output) / (fabs(_mlp_output) + fabs(_new[0]-_old[0])))));
   }
//printf(__FUNCSIG__ + " output is: %.5f, change is: %.5f, and result is: %i", _mlp_output, m_symbol.Bid()-_mlp_output, result);return(0);
   return(result);
}
//+------------------------------------------------------------------+
//| "Voting" that price will fall.                                   |
//+------------------------------------------------------------------+
int CSignalCMLP::ShortCondition(void)
{  int result = 0;
   double _mlp_output = 0.0;
   GetOutput(_mlp_output);
   if(_mlp_output < 0.5)
   {  vector _new, _old;
      _new.CopyRates(m_symbol.Name(), m_period, 8, 0, __MLP_INPUTS);
      _old.CopyRates(m_symbol.Name(), m_period, 8, 1, __MLP_INPUTS);
      result = int(round(100.0 * (fabs(_mlp_output) / (fabs(_mlp_output) + fabs(_new[0]-_old[0])))));
   }
//printf(__FUNCSIG__ + " output is: %.5f, change is: %.5f, and result is: %i", _mlp_output, m_symbol.Bid()-_mlp_output, result);return(0);
   return(result);
}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSignalCMLP::GetOutput(double &Output)
{  Output = 0.0;
   double _learning_rate = m_learning_rate;
   for(int i = m_epochs; i >= 1; i--)
   {  if(m_learning_type == LEARNING_STEP_DECAY)
      {  int _epoch_index = int(MathFloor((m_epochs - i) / m_decay_epoch_steps));
         _learning_rate = m_learning_rate * pow(m_decay_rate_a, _epoch_index);
      }
      else if(m_learning_type == LEARNING_EXPONENTIAL_DECAY)
      {  _learning_rate = m_learning_rate * exp(-1.0 * m_decay_rate_a * (m_epochs - i + 1));
      }
      else if(m_learning_type == LEARNING_POLYNOMIAL_DECAY)
      {  _learning_rate = m_learning_rate * pow(1.0 - ((m_epochs - i) / m_epochs), m_polynomial_power);
      }
      else if(m_learning_type == LEARNING_INVERSE_TIME_DECAY)
      {  _learning_rate = m_prior_learning_rate / (1.0 + (m_decay_rate_a * (m_epochs - i)));
         m_prior_learning_rate = _learning_rate;
      }
      else if(m_learning_type == LEARNING_COSINE_ANNEALING)
      {  _learning_rate = m_min_learning_rate + (0.5 * (m_learning_rate - m_min_learning_rate) * (1.0 + MathCos(((m_epochs - i) * M_PI) / m_epochs)));
      }
      else if(m_learning_type == LEARNING_CYCLICAL)
      {  double _x = fabs(((2.0 * fmod(m_epochs - i, m_epochs)) / m_epochs) - 1.0);
         _learning_rate = m_min_learning_rate + ((m_learning_rate - m_min_learning_rate) * fmax(0.0, (1.0 - _x)));
      }
      else if(m_learning_type == LEARNING_ONE_CYCLE)
      {  double _cycle_position = (double)((m_epochs - i) % (2 * m_epochs)) / (2.0 * m_epochs);
         if (_cycle_position <= 0.5)
         {  _learning_rate = m_min_learning_rate + (2.0 * _cycle_position * (m_learning_rate - m_min_learning_rate));
         }
         else
         {  _learning_rate = m_learning_rate - (2.0 * (_cycle_position - 0.5) * (m_learning_rate - m_min_learning_rate));
         }
      }
      for(int ii = m_train_set; ii >= 0; ii--)
      {  vector _in, _in_new, _in_old, _out, _out_new, _out_old;
         if
         (
         _in_new.Init(__MLP_INPUTS) && 
         _in_new.CopyRates(m_symbol.Name(), m_period, 8, ii + __MLP_OUTPUTS, __MLP_INPUTS) && 
         _in_new.Size() == __MLP_INPUTS &&
         _in_old.Init(__MLP_INPUTS) && 
         _in_old.CopyRates(m_symbol.Name(), m_period, 8, ii + __MLP_OUTPUTS + 1, __MLP_INPUTS) && 
         _in_old.Size() == __MLP_INPUTS
         )
         {  _in = _in_new - _in_old;
            Normalize(_in);
            MLP.Set(_in);
            MLP.Forward();
            if(ii > 0)// train
            {  if
               (
               _out_new.Init(__MLP_OUTPUTS) && 
               _out_new.CopyRates(m_symbol.Name(), m_period, 8, ii, __MLP_OUTPUTS) && 
               _out_new.Size() == __MLP_OUTPUTS &&
               _out_old.Init(__MLP_OUTPUTS) && 
               _out_old.CopyRates(m_symbol.Name(), m_period, 8, ii + 1, __MLP_OUTPUTS) && 
               _out_old.Size() == __MLP_OUTPUTS
               )
               {  _out = _out_new - _out_old;
                  MLP.Get(_out);
                  if(m_learning_type == LEARNING_ADAPTIVE && m_adaptive_type != ADAPTIVE_NONE)
                  {  MLP.Backward(_learning_rate, m_adaptive_type, m_decay_rate_a, m_decay_rate_b, i);
                  }
                  else
                  {  MLP.Backward(_learning_rate, ADAPTIVE_NONE);
                  }
               }
            }
            else if(ii == 0 && i == 1)
            {  Output = MLP.output[0];
            }
         }
      }
   }
}
//+------------------------------------------------------------------+
//| Weights & Biases exporting function for network                  |
//+------------------------------------------------------------------+
void CSignalCMLP::SetOutput(void)
{  if(AccountInfoDouble(ACCOUNT_EQUITY) - m_train_deposit > m_train_criteria && m_train_criteria >= 0.0)
   {  MLP.Set(m_symbol.Name() + "_" + EnumToString(m_period), AccountInfoDouble(ACCOUNT_EQUITY) - m_train_deposit);
   }
}
//+------------------------------------------------------------------+
//| 
//+------------------------------------------------------------------+
void CSignalCMLP::Normalize(vector &Data)
{  vector _copy;
   _copy.Copy(Data);
   if(_copy.Max() - _copy.Min() > 0.0)
   {  for (int i = 0; i < int(Data.Size()); i++)
      {  if(_copy[i] >= 0.0)
         {  Data[i] = 0.5 + (0.5 * ((_copy[i] - _copy.Min()) / (_copy.Max() - _copy.Min())));
         }
         else if(_copy[i] < 0.0)
         {  Data[i] = (0.5 * ((_copy[i] - _copy.Min()) / (_copy.Max() - _copy.Min())));
         }
      }
   }
   else
   {  Data.Fill(0.5);
   }
}
//+------------------------------------------------------------------+
